# -*- coding: utf-8; -*-
#
# Copyright (c) 2016 Álan Crístoffer
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import inspect
import ahio

from multiprocessing import Pipe
from threading import Thread

from moirai.abstract_process_handler import AbstractProcessHandler
from moirai.decorators import decorate_all_methods, dont_raise
from moirai.hardware.cmd_processor import CommandProcessor


def main(pipe):
    """
    Entry point for this process.
    """
    handler = ProcessHandler(pipe)
    handler.run()


@decorate_all_methods(dont_raise)
class ProcessHandler(AbstractProcessHandler):
    """
    Manages this processes' lifecycle and handles IPC.
    """

    def __init__(self, pipe):
        self.cmd_processor = CommandProcessor(self)
        super().__init__('Hardware', pipe)

    def quit(self):
        pass

    def process_command(self, sender, cmd, args):
        """
        Redirects the commands received to the CommandProcessor class.
        """
        if cmd:
            self.cmd_processor.process_command(sender, cmd, args)

    def loop(self):
        pass


def arguments_of(func):
    """
    @returns the list of arguments of func
    """
    def v(name):
        if ps[name].default == inspect.Parameter.empty:
            return {'name': name}
        else:
            return {'name': name, 'default_value': ps[name].default}
    ps = inspect.signature(func).parameters
    return [v(name) for name in ps]


class Hardware(object):
    def list_drivers(self):
        """
        @returns a list of drivers available in this platform.
        """
        return ahio.list_available_drivers()

    def driver_has_setup(self, driver_name):
        """
        @returns True if the driver has a setup function.
        """
        if driver_name not in ahio.list_available_drivers():
            return False
        with ahio.new_driver(driver_name) as driver:
            return hasattr(driver, 'setup')

    def driver_setup_arguments(self, driver):
        """
        @returns a list of field names and default values, ordered.

            {
                "name": "...",
                ["default_value": any]
            }
        """
        if driver not in ahio.list_available_drivers():
            return None
        with ahio.new_driver(driver) as driver:
            if hasattr(driver, 'setup'):
                return arguments_of(driver.setup)
            else:
                return []

    def driver_ports(self, driver):
        if driver not in ahio.list_available_drivers():
            return []
        driver = ahio.new_driver(driver)
        try:
            return driver.available_pins()
        except:
            return []



#!/usr/bin/env python3
# -*- coding: utf-8; -*-
#
# Copyright (c) 2016 Álan Crístoffer
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import codecs
from os import path

import moirai
from setuptools import find_packages, setup

PWD = path.abspath(path.dirname(__file__))
with codecs.open(path.join(PWD, 'README.md'), encoding='utf-8') as f:
    LONG_DESCRIPTION = f.read()


setup(name='moirai',
      version=moirai.__version__,
      description='Digital Control Manager Backend',
      long_description=LONG_DESCRIPTION,
      # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
      classifiers=[
          'Development Status :: 4 - Beta',
          'Intended Audience :: Education',
          'Intended Audience :: Science/Research',
          'License :: OSI Approved :: MIT License',
          'Natural Language :: English',
          'Operating System :: OS Independent',
          'Programming Language :: Python :: 3 :: Only',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',
          'Topic :: Scientific/Engineering :: Human Machine Interfaces'
      ],
      keywords='digital control manager',
      author='Álan Crístoffer',
      author_email='acristoffers@gmail.com',
      url='https://www.github.com/acristoffers/moirai',
      packages=find_packages(),
      license="MIT",
      install_requires=[
          'appdirs',
          'ahio',
          'Flask',
          'pymongo',
          'python-dateutil',
          'numpy',
          'scipy'
      ],
      entry_points={
          'console_scripts': [
              'moirai = moirai.moirai:start'
          ]
      })



# -*- coding: utf-8; -*-
#
# Copyright (c) 2016 Álan Crístoffer
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import ahio
import datetime

from moirai.database import DatabaseV1
from moirai.hardware.configured_hardware import ConfiguredHardware
from moirai.hardware.timer import Timer


class SystemResponseTest(object):
    def __init__(self, test_id):
        self.db = DatabaseV1()
        tests = self.db.get_setting('system_response_tests')
        self.test = next((t for t in tests if t['id'] == test_id), None)
        if self.test is None:
            raise Exception('Test not found')
        self.hardware = ConfiguredHardware()
        configuration = self.db.get_setting('hardware_configuration')
        self.locks = [self.interlock(l) for l in configuration['interlocks']]

    def interlock(self, lock):
        code = compile('y=%s' % lock['expression'], '_string_', 'exec')

        def f(code=code, lock=lock):
            value = self.hardware.read(lock['sensor'])
            scope = {'x': value}
            exec(code, None, scope)
            if scope['y']:
                self.hardware.write(lock['actuator'], lock['actuatorValue'])
                raise Exception('Interlock')
        return f

    def run(self):
        run_time = self.test['points'][-1]['x']
        interval = self.test['logRate']
        t = Timer(run_time, interval)
        port = self.test['output']
        start_time = datetime.datetime.utcnow()
        last_port_value = 0

        self.db.set_setting('current_test', self.test['name'])

        for o in self.test['fixedOutputs']:
            self.hardware.write(o['alias'], o['value'])

        try:
            while self.db.get_setting('current_test') is not None:
                t.sleep()

                for lock in self.locks:
                    lock()

                for sensor in self.test['inputs']:
                    value = self.hardware.read(sensor)
                    self.db.save_test_sensor_value(
                        self.test['name'],
                        sensor,
                        value,
                        t.elapsed(),
                        start_time)

                for point in self.test['points']:
                    if t.elapsed() < point['x']:
                        self.hardware.write(port, point['y'])
                        last_port_value = point['y']
                        self.db.save_test_sensor_value(
                            self.test['name'],
                            port,
                            point['y'],
                            t.elapsed(),
                            start_time)
                        break
        except Exception as e:
            print(e)
            self.db.set_setting('test_error', str(e))

        self.db.save_test_sensor_value(
            self.test['name'],
            port,
            last_port_value,
            t.elapsed(),
            start_time)

        for o in self.test['afterOutputs']:
            self.hardware.write(o['alias'], o['value'])

        self.db.set_setting('current_test', None)



# -*- coding: utf-8; -*-
#
# Copyright (c) 2016 Álan Crístoffer
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import ahio
import datetime

import numpy as np
import math

from moirai.database import DatabaseV1
from moirai.hardware.configured_hardware import ConfiguredHardware
from moirai.hardware.timer import Timer


class Controller(object):
    def __init__(self, controller_id):
        self.db = DatabaseV1()
        cs = self.db.get_setting('controllers')
        self.cs = next((c for c in cs if c['id'] == controller_id), None)
        if self.cs is None:
            raise Exception('Controller not found')
        self.hardware = ConfiguredHardware()
        configuration = self.db.get_setting('hardware_configuration')
        self.locks = [self.interlock(l) for l in configuration['interlocks']]

    def interlock(self, lock):
        code = compile('y=%s' % lock['expression'], '_string_', 'exec')

        def f(code=code, lock=lock):
            value = self.hardware.read(lock['sensor'])
            scope = {'x': value}
            exec(code, None, scope)
            if scope['y']:
                self.hardware.write(lock['actuator'], lock['actuatorValue'])
                raise Exception('Interlock')
        return f

    def run(self):
        run_time = int(self.cs['runTime'])
        interval = float(self.cs['tau'])
        t = Timer(run_time, interval)
        start_time = datetime.datetime.utcnow()

        self.db.set_setting('current_test', self.cs['name'])

        after = None

        try:
            before = compile(self.cs['before'], '_string_', 'exec')
            controller = compile(self.cs['controller'], '_string_', 'exec')
            after = compile(self.cs['after'], '_string_', 'exec')

            inputs = {s: self.hardware.read(s) for s in self.cs['inputs']}
            plocals = {
                'inputs': inputs,
                'outputs': dict(),
                's': dict()
            }
            pglobals = {
                'np': np,
                'math': math
            }
            exec(before, pglobals, plocals)

            state = plocals['s']

            for k, v in plocals['outputs'].items():
                self.hardware.write(k, v)

            while self.db.get_setting('current_test') is not None:
                t.sleep()
                time = t.elapsed()

                for lock in self.locks:
                    lock()

                inputs = {s: self.hardware.read(s) for s in self.cs['inputs']}
                plocals = {
                    'inputs': inputs,
                    'outputs': dict(),
                    's': state,
                    'log': dict()
                }
                pglobals = {
                    'np': np,
                    'math': math
                }
                exec(controller, pglobals, plocals)

                plocals['log'] = {
                    **plocals['log'],
                    **plocals['inputs'],
                    **plocals['outputs']
                }

                for k, v in plocals['outputs'].items():
                    self.hardware.write(k, v)

                for k, v in plocals['log'].items():
                    cid = self.cs['name']
                    self.db.save_test_sensor_value(cid, k, v, time, start_time)

                state = plocals['s']
        except Exception as e:
            print(e)
            self.db.set_setting('test_error', str(e))

        if after is not None:
            inputs = {s: self.hardware.read(s) for s in self.cs['inputs']}
            plocals = {
                'inputs': inputs,
                'outputs': dict()
            }
            pglobals = {
                'np': np,
                'math': math
            }
            exec(after, pglobals, plocals)

            for actuator, value in plocals['outputs'].items():
                self.hardware.write(actuator, value)

        self.db.set_setting('current_test', None)
